%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% set_sw/1,set_sw/2: initialize the prob. of MSW
%% examples: set_sw(i,0.6+0.4)
%%           set_sw(i,[0.6,0.4])
%%           set_sw(i,uniform)
%%           set_sw(i,f_geometric)
%%           set_sw(i,f_geometric(2))
%%           set_sw(i,f_geometric(2,asc))
%%           set_sw(i)   % uniform distribution

set_sw(Sw) :- set_sw(Sw,default).

set_sw(Sw,Dist) :-
    $pp_require_ground(Sw,$msg(0101),set_sw/2),
    $pp_require_switch_outcomes(Sw,$msg(0102),set_sw/2),
    $pp_require_distribution(Dist,$msg(0200),set_sw/2),
    $pp_set_sw(Sw,Dist).

$pp_set_sw(Sw,Dist) :-
    ( $pd_fixed_prob_distrib(Sw) -> $pp_raise_warning($msg(0109),[Sw])
    ; $pp_get_values(Sw,Values),
      length(Values,N),
      dist_to_probs(Dist,N,Probs),
      ( retract($pd_prob_distrib(Sw,_,_)) -> true ; true),
      assert($pd_prob_distrib(Sw,Values,Probs))
    ),!.

%% set_sw_all(Sw): set parameters to all switches that matches with Sw.

set_sw_all          :- $pp_set_sw_all(_,default).
set_sw_all(Sw)      :- $pp_set_sw_all(Sw,default).
set_sw_all(Sw,Dist) :- $pp_set_sw_all(Sw,Dist).

$pp_set_sw_all(Sw,Dist) :-
    findall(Sw,$pd_registered_sw(Sw),Sws0),
    $pp_set_sw_list(Sws0,Dist),!,
    findall(Sw,(get_values2(Sw,_),ground(Sw)),Sws1),
    $pp_set_sw_list(Sws1,Dist),!.

$pp_set_sw_list([],_).
$pp_set_sw_list([Sw|Sws],Dist) :-
    set_sw(Sw,Dist),!,
    $pp_set_sw_list(Sws,Dist).

% fix switches
fix_sw(Sw,Params) :-
    set_sw(Sw,Params),
    $pp_fix_sw(Sw).

fix_sw(Sw) :- var(Sw),!,
    ( get_sw(switch(Sw1,_,_,_)),
      fix_sw(Sw1),
      fail
    ; true
    ).
fix_sw(Sw) :- Sw = [_|_],!,
    $pp_fix_sw_list(Sw).
fix_sw(Sw) :-
    ( $pd_prob_distrib(Sw,_,_),
      $pp_fix_sw(Sw),
      fail
    ; true
    ).

$pp_fix_sw_list([]).
$pp_fix_sw_list([Sw|Sws]) :-
    fix_sw(Sw),!,
    $pp_fix_sw_list(Sws).

$pp_fix_sw(Sw) :-
    ( $pd_fixed_prob_distrib(Sw) -> true
    ; assert($pd_fixed_prob_distrib(Sw))
    ).

unfix_sw(Sw) :- var(Sw),!,
    ( get_sw(switch(Sw1,_,_,_)),
      unfix_sw(Sw1),
      fail
    ; true
    ).
unfix_sw(SwList) :- SwList = [_|_],!,$pp_unfix_sw_list(SwList).
unfix_sw(Sw) :-
    ( $pd_prob_distrib(Sw,_,_),
      $pp_unfix_sw(Sw),
      fail
    ; true
    ).

$pp_unfix_sw_list([]).
$pp_unfix_sw_list([Sw|Sws]) :-
    $pp_unfix_sw(Sw),!,
    $pp_unfix_sw_list(Sws).

$pp_unfix_sw(Sw) :-
    ( retract($pd_fixed_prob_distrib(Sw)) -> true ; true).

% show msw
show_sw :- show_sw(_).

show_sw(Sw) :-
   findall(Sw,$pd_prob_distrib(Sw,_,_),Sws0),
   sort(Sws0,Sws),
   $pp_show_sw_list(Sws).

$pp_show_sw_list([]) :- !.
$pp_show_sw_list([Sw|Sws]) :-!,
   $pp_show_sw1(Sw),!,
   $pp_show_sw_list(Sws).

% We can assume Sw is ground
$pp_show_sw1(Sw) :-
    $pp_get_distribution(Sw,Values,Probs),
    format("Switch ~w: ",[Sw]),
    ( $pd_fixed_prob_distrib(Sw) -> write('fixed_p:') ; write('unfixed_p:') ),
    $pp_show_sw_values(Values,Probs),
    nl.

$pp_show_sw_values([],_Ps).
$pp_show_sw_values([V|Vs],[P|Ps]) :-
    format(" ~w (p: ~9f)",[V,P]),!,
    $pp_show_sw_values(Vs,Ps).

get_sw(Sw) :-
    get_sw(SwName,Status,Values,Probs),
    Sw = switch(SwName,Status,Values,Probs).

get_sw(Sw,[Status,Values,Probs]) :-
    get_sw(Sw,Status,Values,Probs).

% - Since $pp_get_values/2 is not backtrackable, we should call
%   $pd_prob_distrib/3 in advance for the case in which Sw is not ground.
%
% - Inconsitency of outcome spaces are checked in advance in $pp_get_distribution/3
%   and $pp_get_expectations/3.

get_sw(Sw,Status,Values,Probs) :-
    $pp_get_distribution(Sw,Values,Probs),
    $pp_get_values(Sw,_),
    ( $pd_fixed_prob_distrib(Sw) -> Status = fixed ; Status = unfixed ).
    
get_sw(Sw,Status,Values,Probs,Es) :-
    $pp_get_distribution(Sw,Values,Probs),
    $pp_get_expectations(Sw,_,Es),
    $pp_get_values(Sw,_),
    ( $pd_fixed_prob_distrib(Sw) -> Status = fixed ; Status = unfixed ).

%% save/restore switch information

save_sw :- save_sw('Saved_SW').

save_sw(File) :-
    open(File,write,OutStream),
    ( get_sw(SwName,Status,Values,Probs),
      format(OutStream,"switch(~q,~q,~q,",[SwName,Status,Values]),
      $pp_write_distribution(OutStream,Probs,'['),
      format(OutStream,"]).~n",[]),
      fail
    ; true
    ),
    close(OutStream),!.

$pp_write_distribution(_,[],_).
$pp_write_distribution(OutStream,[Prob|Probs],C) :-
    format(OutStream,"~w~15e",[C,Prob]),!,
    $pp_write_distribution(OutStream,Probs,',').

restore_sw :- restore_sw('Saved_SW').

restore_sw(File) :-
    open(File,read,InStream),
    repeat,
    read(InStream,Switch),
    ( Switch == end_of_file
    ; Switch = switch(ID,_,_,Params),
      set_sw(ID,Params),
      fail
    ),
    close(InStream),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% set_sw_h/1,set_sw_h/2: initialize the pseudo counts of MSW
% examples: set_sw_h(i,[1.0,2.0])
%           set_sw_h(i,3.0)          ... all pseudo counts for i are set to 3.0
%           set_sw_h(i,uniform)      ... all pseudo counts for i are set to 1.0/|V_i|
%           set_sw_h(i,uniform(3.0)) ... all pseudo counts for i are set to 3.0/|V_i|
%           set_sw_h(i)              ... the same as set_sw_h(i,uniform)

set_sw_h(Sw) :- set_sw_h(Sw,default).

set_sw_h(Sw,Spec) :-
    $pp_require_ground(Sw,$msg(0101),set_sw_h/2),
    $pp_require_switch_outcomes(Sw,$msg(0102),set_sw_h/2),
    $pp_require_pseudo_count_spec(Spec,$msg(0201),set_sw_h/2),
    $pp_set_sw_h(Sw,Spec).

$pp_set_sw_h(Sw,Spec) :-
    ( $pd_fixed_pseudo_counts(Sw) -> $pp_raise_warning($msg(0110),[Sw])
    ; $pp_get_values(Sw,Values),
      length(Values,N),
      spec_to_pseudo_counts(Spec,N,PseudoCs),
      ( retract($pd_pseudo_counts(Sw,_,_)) -> true ; true),
      assert($pd_pseudo_counts(Sw,Values,PseudoCs))
    ),!.

%%% set_sw_all_h(Sw): set hyparameters to all switches that matches with Sw.

set_sw_all_h :- set_sw_all_h(_).

set_sw_all_h(Sw) :- set_sw_all_h(Sw,default).

set_sw_all_h(Sw,PseudoCs) :-
    findall(Sw,$pp_registered_sw(Sw),Sws0),
    $pp_set_sw_h_list(Sws0,PseudoCs),!,
    findall(Sw,(get_values2(Sw,_),ground(Sw)),Sws1),
    $pp_set_sw_h_list(Sws1,PseudoCs),!.

$pp_set_sw_h_list([],_).
$pp_set_sw_h_list([Sw|Sws],PseudoCs) :-
    set_sw_h(Sw,PseudoCs),!,
    $pp_set_sw_h_list(Sws,PseudoCs).

% aliases
set_sw_h_all              :- set_sw_all_h.
set_sw_h_all(Sw)          :- set_sw_all_h(Sw).
set_sw_h_all(Sw,PseudoCs) :- set_sw_all_h(Sw,PseudoCs).

%%% fix_sw_h(Sw,PseudoCs) :- fix the pseudo counts of Sw at PseudoCs.

fix_sw_h(Sw,Params) :-
    set_sw_h(Sw,Params),
    $pp_fix_sw_h(Sw).

fix_sw_h(Sw) :- var(Sw),!,
    ( get_sw_h(switch(Sw1,_,_,_)),
      fix_sw_h(Sw1),
      fail
    ; true
    ).

fix_sw_h(Sw) :- Sw = [_|_],!,
    $pp_fix_sw_h_list(Sw).
fix_sw_h(Sw) :-
    ( $pd_pseudo_counts(Sw,_,_),
      $pp_fix_sw_h(Sw),
      fail
    ; true
    ),!.

$pp_fix_sw_h_list([]).
$pp_fix_sw_h_list([Sw|Sws]) :-
    fix_sw_h(Sw),!,
    $pp_fix_sw_h_list(Sws).

$pp_fix_sw_h(Sw) :-
    ( clause($pd_fixed_pseudo_counts(Sw),_) -> true
    ; assert($pd_fixed_pseudo_counts(Sw))
    ).

unfix_sw_h(Sw) :- var(Sw),!,
    ( get_sw_h(switch(Sw1,_,_,_)),
      unfix_sw_h(Sw1),
      fail
    ; true
    ).
unfix_sw_h(SwList) :- SwList = [_|_],!,$pp_unfix_sw_h_list(SwList).
unfix_sw_h(Sw) :-
    ( $pd_pseudo_counts(Sw,_,_),
      $pp_unfix_sw_h(Sw),
      fail
    ; true
    ),!.

$pp_unfix_sw_h_list([]).
$pp_unfix_sw_h_list([Sw|Sws]) :-
    unfix_sw_h(Sw),!,
    $pp_unfix_sw_h_list(Sws).

$pp_unfix_sw_h(Sw) :-
    ( retract($pd_fixed_pseudo_counts(Sw)) -> true
    ; true
    ).

%%% show pseudo counts

show_sw_h :- show_sw_h(_).

show_sw_h(Sw) :-
   findall(Sw,$pd_pseudo_counts(Sw,_,_),Sws0),
   sort(Sws0,Sws),
   $pp_show_sw_list_h(Sws).

$pp_show_sw_list_h([]) :- !.
$pp_show_sw_list_h([Sw|Sws]) :- !,
   $pp_show_sw1_h(Sw),!,
   $pp_show_sw_list_h(Sws).

$pp_show_sw1_h(Sw) :-
    $pp_get_pseudo_counts(Sw,Values,PseudoCs),
    format("Switch ~w: ",[Sw]),
    ( $pd_fixed_pseudo_counts(Sw) -> write('fixed_h:') ; write('unfixed_h:') ),
    $pp_show_sw_h_values(Values,PseudoCs),
    nl.

$pp_show_sw_h_values([],_).
$pp_show_sw_h_values([V|Vs],[H|Hs]) :-
    format(" ~w (c: ~9f)",[V,H]),!,
    $pp_show_sw_h_values(Vs,Hs).

%%% show both parameters and pseudo counts

show_sw_b :- show_sw_b(_).

show_sw_b(Sw) :-
    findall(Sw,$pp_registered_sw(Sw),Sws0),
    sort(Sws0,Sws),
    $pp_show_sw_list_b(Sws).

$pp_show_sw_list_b([]) :- !.
$pp_show_sw_list_b([Sw|Sws]) :- !,
    $pp_show_sw1_b(Sw),!,
    $pp_show_sw_list_b(Sws).

$pp_show_sw1_b(Sw) :-
    $pp_get_distribution(Sw,Values,Probs),
    $pp_get_pseudo_counts(Sw,_,PseudoCs),
    format("Switch ~w: ",[Sw]),
    ( $pd_fixed_prob_distrib(Sw)  -> write('fixed_p,') ; write('unfixed_p,') ),
    ( $pd_fixed_pseudo_counts(Sw) -> write('fixed_h:') ; write('unfixed_h:') ),
    $pp_show_sw_b_values(Values,Probs,PseudoCs),
    nl,!.

$pp_show_sw_b_values([],_,_).

$pp_show_sw_b_values([V|Vs],[P|Ps],[SC|SCs]) :-
    format(" ~w (p: ~9f, c: ~9f)",[V,P,SC]),!,
    $pp_show_sw_b_values(Vs,Ps,SCs).

$pp_show_sw_b_values([V|Vs],[P|Ps],$not_assigned) :-
    format(" ~w (p: ~9f, c: n/a)",[V,P]),!,
    $pp_show_sw_b_values(Vs,Ps,$not_assigned).

$pp_show_sw_b_values([V|Vs],$not_assigned,[SC|SCs]) :-
    format(" ~w (p: n/a, c: ~9f)",[V,SC]),!,
    $pp_show_sw_b_values(Vs,$not_assigned,SCs).

%%% get switch information including pseudo counts

get_sw_h(Sw) :-
    get_sw_h(SwName,Status,Values,PseudoCs),
    Sw = switch(SwName,Status,Values,PseudoCs).

get_sw_h(Sw,[Status,Values,PseudoCs]) :- get_sw_h(Sw,Status,Values,PseudoCs).

% [Note] Since $pp_get_values/2 is deterministic, we should call
% $pp_get_pseudo_counts/3 in advance for the case in which Sw is not ground.
get_sw_h(Sw,Status,Values,PseudoCs) :-
    $pp_get_pseudo_counts(Sw,Values,PseudoCs),
    $pp_get_values(Sw,_),
    ( $pd_fixed_pseudo_counts(Sw) -> Status = fixed ; Status = unfixed ).

%%% get switch information including both parameters and pseudo counts

get_sw_b(Sw) :-
    get_sw_b(SwName,StatusPair,Values,Probs,PseudoCs),
    Sw = switch(SwName,StatusPair,Values,Probs,PseudoCs).

get_sw_b(Sw,[StatusPair,Values,Probs,PseudoCs]) :-
    get_sw_b(Sw,StatusPair,Values,Probs,PseudoCs).

get_sw_b(Sw,[StatusP,StatusH],Values,Probs,PseudoCs) :-
    $pp_get_distribution(Sw,Values,Probs),
    $pd_get_pseudo_counts(Sw,_,PseudoCs),
    $pp_get_values(Sw,_),
    ( $pd_fixed_prob_distrib(Sw)  -> StatusP = fixed ; StatusP = unfixed ),
    ( $pd_fixed_pseudo_counts(Sw) -> StatusH = fixed ; StatusH = unfixed ).

get_sw_b(Sw,[StatusP,StatusH],Values,Probs,Es,PseudoCs) :-
    $pp_get_distribution(Sw,Values,Probs),
    $pd_get_pseudo_counts(Sw,_,PseudoCs),
    $pp_get_expectations(Sw,_,Es),
    $pp_get_values(Sw,_),
    ( $pd_fixed_prob_distrib(Sw)  -> StatusP = fixed ; StatusP = unfixed ),
    ( $pd_fixed_pseudo_counts(Sw) -> StatusH = fixed ; StatusH = unfixed ).

%% save/restore pseudo counts

save_sw_h :- save_sw_h('Saved_SW_H').

save_sw_h(File) :-
    open(File,write,OutStream),
    ( get_sw_h(SwName,Status,Values,PseudoCs),
      format(OutStream,"switch(~q,~q,~q,",[SwName,Status,Values]),
      $pp_write_pseudo_counts(OutStream,PseudoCs,'['),
      format(OutStream,"]).~n",[]),
      fail
    ; true
    ),
    close(OutStream),!.

$pp_write_pseudo_counts(_,[],_).
$pp_write_pseudo_counts(OutStream,[PseudoC|PseudoCs],C) :-
    format(OutStream,"~w~15e",[C,PseudoC]),!,
    $pp_write_pseudo_counts(OutStream,PseudoCs,',').

restore_sw_h :- restore_sw_h('Saved_SW_H').

restore_sw_h(File) :-
    open(File,read,InStream),
    repeat,
    read(InStream,Switch),
    ( Switch == end_of_file
    ; Switch = switch(ID,_,_,PseudoCs),
        set_sw_h(ID,PseudoCs),fail
    ),
    close(InStream),!.

%% save/restore both parameters and pseudo counts

save_sw_b :- save_sw, save_sw_h.

save_sw_b(FileP,FileH) :-
    save_sw(FileP),
    save_sw_h(FileH),!.

restore_sw_b :- restore_sw, restore_sw_h.

restore_sw_b(FileP,FileH) :-
    restore_sw(FileP),
    restore_sw_h(FileH),!.

%%----------------------------------------
%%  [Note]
%%  $pp_get_{values,distribution,expectations,pseudo_counts}/2 do not check
%%  the groundness of switch names.

% raises a exception if some error occurs
get_values(Sw,Values) :-
     $pp_require_ground(Sw,$msg(0101),get_values/2),
     $pp_get_values(Sw,Values).

% just fails if some error occurs; Sw should be ground
get_values1(Sw,Values) :-
     ground(Sw),
     current_predicate(values/2),
     values(Sw,Values).

% just fails if some error occurs; Sw doesn't have to be ground
get_values2(Sw,Values) :-
     catch(values(Sw,Values),_,fail).

% Sw is assumed to be ground
$pp_get_values(Sw,Values) :-
     $pp_require_switch_outcomes(Sw,$msg(0102),$pp_get_values/2),
     values(Sw,Values).

$pp_get_distribution(Sw,Values,Probs) :-
     ( ground(Sw) ->
	 $pp_get_values(Sw,Values),
         ( $pd_prob_distrib(Sw,Values0,Probs0) ->
	     ( Values0 = Values -> Probs = Probs0
	     ; % if the outcome space has been changed, reset the distribution:
               get_prism_flag(dynamic_default_sw,on) ->
                 ( $pd_fixed_prob_distrib(Sw) ->
                     $pp_raise_runtime_error($msg(0106),[Sw],
                                             modified_switch_outcomes,
                                             $pp_get_distribution/3)
                 ; set_sw(Sw,default),!,
                   $pd_prob_distrib(Sw,Values,Probs)
                 )
	     ; $pp_raise_runtime_error($msg(0105),[Sw],
                                       modified_switch_outcomes,
                                       $pp_get_distribution/3)
	     )
         ; set_sw(Sw,default),!,
           $pd_prob_distrib(Sw,Values,Probs)
         )
     ; $pd_prob_distrib(Sw,Values,Probs)
         % if Sw is not ground, we do not assign the default distribution
     ).

$pp_get_expectations(SW,Values,Es) :-
     ( ground(Sw) ->
         $pp_get_values(Sw,Values),
         $pd_expectation(SW,Values0,Es0),
         ( Values0 = Values -> Es = Es0
         ; $pp_raise_runtime_error($msg(0107),[Sw],modified_switch_outcomes,
                                   $pp_get_expectations/3)
         )
     ; $pd_expectation(SW,Values,Es)
     ).

$pp_get_pseudo_counts(Sw,Values,PseudoCs) :-
     ( ground(Sw) ->
         $pp_get_values(Sw,Values),
         ( $pd_pseudo_counts(Sw,Values0,PseudoCs0) ->
             ( Values0 = Values -> PseudoCs = PseudoCs0
             ; get_prism_flag(dynamic_default_sw_h,on) ->
                 ( $pd_fixed_pseudo_counts(Sw) ->
                     $pp_raise_runtime_error($msg(0108),[Sw],
                                             modified_switch_outcomes,
                                             $pp_get_pseudo_counts/3)
                 ; set_sw_h(Sw,default),!,
                   $pd_pseudo_counts(Sw,Values,PseudoCs)
                 )
             ; $pp_raise_runtime_error($msg(0105),[Sw],
                                       modified_switch_outcomes,
                                       $pp_get_pseudo_counts/3)
             )
         ; set_sw_h(Sw,default),!,
           $pd_pseudo_counts(Sw,Values,PseudoCs)
         )
     ; $pd_pseudo_counts(Sw,Values,PseudoCs)
     ).

%%----------------------------------------

$pp_registered_sw(Sw) :-   % ground switch name will be returned
     ( $pd_prob_distrib(Sw,_,_)
     ; $pd_pseudo_counts(Sw,_,_)
     ).

show_reg_sw :-
     get_reg_sw(Sws),
     $pp_show_reg_sw(Sws).

$pp_show_reg_sw(Sws) :-
     format("Registered random switches:~n",[]),
     $pp_show_reg_sw1(Sws).

$pp_show_reg_sw1([]).
$pp_show_reg_sw1([Sw|Sws]) :-
     format("  ~w~n",[Sw]),!,
     $pp_show_reg_sw1(Sws).

get_reg_sw(Sws) :-
     findall(Sw,$pp_registered_sw(Sw),Sws0),!,
     sort(Sws0,Sws).
